export = Providers;
/**
 * @typedef {import('multiformats/cid').CID} CID
 * @typedef {import('interface-datastore').Datastore} Datastore
 */
/**
 * This class manages known providers.
 * A provider is a peer that we know to have the content for a given CID.
 *
 * Every `cleanupInterval` providers are checked if they
 * are still valid, i.e. younger than the `provideValidity`.
 * If they are not, they are deleted.
 *
 * To ensure the list survives restarts of the daemon,
 * providers are stored in the datastore, but to ensure
 * access is fast there is an LRU cache in front of that.
 */
declare class Providers {
    /**
     * @param {Datastore} datastore
     * @param {PeerId} [self]
     * @param {number} [cacheSize=256]
     */
    constructor(datastore: Datastore, self?: PeerId | undefined, cacheSize?: number | undefined);
    datastore: import("interface-datastore/dist/src/types").Datastore;
    _log: debug.Debugger & {
        error: debug.Debugger;
    };
    /**
     * How often invalid records are cleaned. (in seconds)
     *
     * @type {number}
     */
    cleanupInterval: number;
    /**
     * How long is a provider valid for. (in seconds)
     *
     * @type {number}
     */
    provideValidity: number;
    /**
     * LRU cache size
     *
     * @type {number}
     */
    lruCacheSize: number;
    providers: any;
    syncQueue: Queue<import("p-queue/dist/priority-queue").default, import("p-queue").DefaultAddOptions>;
    /**
     * Start the provider cleanup service
     */
    start(): void;
    _cleaner: NodeJS.Timer | null | undefined;
    /**
     * Release any resources.
     */
    stop(): void;
    /**
     * Check all providers if they are still valid, and if not delete them.
     *
     * @returns {Promise<void>}
     * @private
     */
    private _cleanup;
    /**
     * Get the currently known provider peer ids for a given CID.
     *
     * @param {CID} cid
     * @returns {Promise<Map<string, Date>>}
     *
     * @private
     */
    private _getProvidersMap;
    /**
     * Add a new provider for the given CID.
     *
     * @param {CID} cid
     * @param {PeerId} provider
     * @returns {Promise<void>}
     */
    addProvider(cid: CID, provider: PeerId): Promise<void>;
    /**
     * Get a list of providers for the given CID.
     *
     * @param {CID} cid
     * @returns {Promise<Array<PeerId>>}
     */
    getProviders(cid: CID): Promise<Array<PeerId>>;
}
declare namespace Providers {
    export { CID, Datastore };
}
import { default as Queue } from "p-queue";
type CID = import('multiformats/cid').CID;
import PeerId = require("peer-id");
type Datastore = import('interface-datastore').Datastore;
//# sourceMappingURL=providers.d.ts.map